package scanner

import error.LineColumnException
import file.FilePosition
import file.FilePositionTracker
import token.Token
import token.TokenFactoryImpl
import util.Characters
import util.Characters.BACKSLASH
import util.Characters.DOUBLE_QUOTE
import util.Characters.EOF
import util.Characters.NEWLINE
import util.Characters.isIdentifierCharacter
import util.Characters.isNumberPrefix
import java.io.InputStream
import java.lang.Character.isDigit
import java.lang.Character.isWhitespace

/**
 * Converts a stream of bytes into a stream of Lisp tokens.
 */
class LispScanner(inputStream: InputStream, fileName: String) {

    private val inputStream = LispCommentRemovingInputStream(inputStream)
    private val positionTracker = FilePositionTracker(fileName)
    private val tokenFactory = TokenFactoryImpl()

    val nextToken: Token
        get() {
            var c = inputStream.read()
            while (c != EOF) {
                val currentCharacter = c.toChar()
                positionTracker.incrementColumn()

                if (!isWhitespace(currentCharacter))
                    return createTokenFromCharacter(currentCharacter)
                else if (currentCharacter == NEWLINE)
                    positionTracker.incrementLine()

                c = inputStream.read()
            }

            return tokenFactory.createEofToken(positionTracker.currentPosition())
        }

    private fun createTokenFromCharacter(c: Char) = positionTracker.currentPosition().let { position ->
        tokenFactory.createToken(retrieveTokenText(c), position)
    }

    private fun retrieveTokenText(firstCharacter: Char): String {
        var tokenText = "" + firstCharacter

        when {
            firstCharacter == DOUBLE_QUOTE        -> tokenText = retrieveStringTokenText(firstCharacter)
            isNumberPrefix(firstCharacter)        -> tokenText = retrieveNumberOrIdentifierTokenText(firstCharacter)
            firstCharacter.isDigit()              -> tokenText = retrieveNumberTokenText(firstCharacter)
            isIdentifierCharacter(firstCharacter) -> tokenText = retrieveIdentifierTokenText(firstCharacter)
        }

        return tokenText
    }

    private fun retrieveNumberOrIdentifierTokenText(firstCharacter: Char): String {
        val nextCharacter = inputStream.read().toChar()
        inputStream.unreadLastCharacter()

        return if (isDigit(nextCharacter))
            retrieveNumberTokenText(firstCharacter)
        else
            retrieveIdentifierTokenText(firstCharacter)
    }

    private fun retrieveStringTokenText(firstDoubleQuote: Char) =
        ComplexTokenTextRetriever(firstDoubleQuote) { true }.retrieveToken()

    private fun retrieveNumberTokenText(firstCharacter: Char) =
        ComplexTokenTextRetriever(firstCharacter, Character::isDigit).retrieveToken()

    private fun retrieveIdentifierTokenText(firstCharacter: Char) =
        ComplexTokenTextRetriever(firstCharacter, Characters::isIdentifierCharacter).retrieveToken()

    private inner class ComplexTokenTextRetriever(internal var firstCharacter: Char,
                                                  internal var isPartOfToken: (Char) -> Boolean) {
        internal var text = StringBuilder()
        internal var position = positionTracker.currentPosition()
        internal var currentCharacter = firstCharacter
        internal var previousCharacter = firstCharacter

        private fun isTerminatingCharacter() = isStringToken() && isTerminatingDoubleQuote()
        private fun isStringToken() = firstCharacter == DOUBLE_QUOTE
        private fun isTerminatingDoubleQuote() = currentCharacter == DOUBLE_QUOTE && previousCharacter != BACKSLASH

        fun retrieveToken(): String {
            text.append(firstCharacter)

            var c = inputStream.read()
            while (c != EOF) {
                currentCharacter = c.toChar()

                if (!isPartOfToken(currentCharacter)) {
                    inputStream.unreadLastCharacter()

                    return text.toString()
                }

                addCharacterToToken()

                if (isTerminatingCharacter())
                    return text.toString()

                previousCharacter = currentCharacter
                c = inputStream.read()
            }

            return terminateTokenWithEof()
        }

        private fun addCharacterToToken() {
            text.append(currentCharacter)
            positionTracker.incrementColumn()

            if (currentCharacter == NEWLINE)
                positionTracker.incrementLine()
        }

        private fun terminateTokenWithEof(): String {
            if (isStringToken())
                throw UnterminatedStringException(position)

            return text.toString()
        }
    }

    class UnterminatedStringException(position: FilePosition) : LineColumnException(position) {
        override val messagePrefix: String
            get() = "unterminated quoted string"
    }
}
